*******************************************************************************;
** SAS macro MR2RM    vs. 2.7.3n, 26 Jul 2012, by Jim Groeneveld, Netherlands *;
*******************************************************************************;
**                                                                            *;
*; %MACRO   MR2RM      /* Conversion of Multiple Records To Repeated Measures */
/* -------------------------------------------------------------------------- */
/* ---------- A l w a y s   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  Data    =_LAST_     /* Input data set specification                      R */
,  Out     =           /* Output data set specification (may be InputData)  R */
,  ByList  =           /* Variable List of which multiple Identical value   R */
                       /*     combinations determine a single new record      */
                       /*     (often variables like a Case Number)            */
/*                                                                            */
/* ------ F r e q u e n t l y   s p e c i f i e d   a r g u m e n t s ------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  OverWrit=N          /* specify Y to overwrite an existing Out dataset    O */
,  VarList =           /* Multiple Records variable name List to convert    O */
                       /*     to Repeated Measures using PreFxLst and Index,  */
                       /*     No value or '_ALL_' involves all variables      */
,  Index   =           /* Indexing measures variable name: numeric integer  O */
                       /*     and unique within ByList, default automatic ~)  */
,  CopyList=           /* list of (remaining) variables to include, but to  O */
                       /*   leave as they are, resulting in single values,    */
                       /*   - empty(none)/varlist/_REST_(not in other lists)  */
/*                                                                            */
/* ---- O c c a s i o n a l l y   s p e c i f i e d   a r g u m e n t s ----- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  ByFormat=UNFORMATTED/* identify cases by unformatted values of identifi- O */
                       /*     cation variables, alternative argument value:   */
                       /*   FORMATTED (or explicitely empty) : identify       */
                       /*     cases by eventually formatted values of         */
                       /*     identification variables                        */
,  Where   =           /* complete where clause, parentheses not necessary  O */
,  PreFxLst=           /* Prefix variable name list for result variables    O */
                       /*     to be postfixed with indexing number `)         */
,  IndPreFx=/          /* Index prefix (separator) text in variable labels, O */
                       /*     avoid special characters like %, &, ', and "    */
,  Lowest  =           /*  Lowest (integer) value of index variable         O */
,  Highest =           /* Highest (integer) value of index variable         O */
/*                                                                            */
/* ---------- R a r e l y   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  IndexLen=Maximum    /* Maximum: index postfix width = width of longest   O */
                       /* Minimum: index postfix width = index digits width   */
,  CopyFrom=Updated    /* where the single kept values are to come from:    O */
                       /*   - LastRec(last MR)/Updated(last non-mis value)    */
,  DupIndex=Exclude    /* or Include duplicate Index values in extra rec ) O */
                       /*   + if Freqvars=Yes then _Record_ is also output    */
,  FreqVars=No         /* Number of passed and missed MRs in output: Y/N ') O */
                       /* (frequency variables are _Passed_ and _Missed_)     */
,  Version = 2.7.3n    /* forced version specification for version control  O */
,  CalledBy=           /* Macro name of macro or program for err0r report   O */
/*                                                                            */
/* ----------- N e v e r   s p e c i f i e d   a r g u m e n t s ------------ */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
                       /* Extinct, but yet supported old version parameters:  */
,  SourceDs=           /* old version's Data    , one specification allowed - */
,  TargetDs=           /* old version's Out     , one specification allowed - */
,  IDlist  =           /* old version's ByList  , one specification allowed - */
,  IdFormat=           /* old version's ByFormat, one specification allowed - */
,  MRlist  =           /* old version's VarList , one specification allowed - */
,  KeepList=           /* old version's CopyList, one specification allowed - */
,  KeepFrom=           /* old version's CopyFrom, one specification allowed - */
                       /* Use only new parameter names with new macro calls!  */
                       /* - - - - - - - - - - - - - - - - - - - - - - - - - - */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  ForMyUse=           /* argument for the author's use only, do not use it ! */
)/ DES     =             'Convert Mult Recs to Repeated Measures' /* STORE */  ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : MR2RM.SAS / MR2RM.273                                         *;
%*                                                                            *;
%* Purpose    : Transformation of datasets with Multiple Responses as single  *;
%*              variables in multiple records (observations) into one with    *;
%*              those as multiple variables (Repeated Measurements) in a      *;
%*              single record.                                                *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 26 Jul 2012                                                   *;
%* Version    : 2.7.3n                                                        *;
%*                                                                            *;
%* SAS version: 6.12 and W95, 8.2 and W95/W98/W2K/WXP, 9.1.3 and WXP          *;
%*                                                                            *;
%* Validation : - (not yet validated, check results, use at your own risk)    *;
%* Disclaimer : The author is not liable in any way for any negative, adverse *;
%*              consequences of the use or misuse of this program / macro.    *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Notes:                                                                     *;
%* ======                                                                     *;
%*                                                                            *;
%* ~)-Generating an index variable automatically should be applied with care: *;
%*    The generated values will be incremental from 1 or Lowest within each   *;
%*    group determined by unique occurring and sorted value combinations of   *;
%*    the identification variables in ByList. If groups differ in size they   *;
%*    also differ with respect to the generated, max index value range within!*;
%*   -A specified index variable may also occur as (part of) a VarList and/or *;
%*    as (part of) an ByList, but the index variable will always be dropped.  *;
%*   -If argument Lowest is specified it determines the starting value of the *;
%*    automatically generated index. Any specified Highest argument determines*;
%*    the maximum (generated) index value for inclusion in the Out dataset.   *;
%*   -RM Variable names are being generated on the basis of the original      *;
%*    ByList variable names and will have characters postfixed, which are     *;
%*    based on the non-missing and missing numeric Index values:              *;
%*    . positive (integer) index values generate just digits (as text) being  *;
%*      postfixed to the original (or possibly truncated) variable names,     *;
%*      padded with leading zeros up to the length of the postfix.            *;
%*    . negative (integer) index values generate a _ char. and digits being   *;
%*      postfixed to the original (or possibly truncated) variable names,     *;
%*      padded with leading zeros up to the length of the postfix.            *;
%*   -Missing index values always generate a repeated measures variable list. *;
%*    Missing index values ._, ., .a - .z generate postfixes _, _, a - z if   *;
%*    one postfix character is sufficient and values ._ and . do not occur    *;
%*    both. If more postfix characters are necessary or available they are    *;
%*    __, _0, _a - _z, extended with leading _ characters to the length of    *;
%*    the postfix.                                                            *;
%*    (Missing ByList values are left as they are and always are included.)   *;
%*                                                                            *;
%* `)-The user should take care of not to cause duplicate variable names and  *;
%*    names that are too long. The specified prefixes should be short enough  *;
%*    to accomodate for valid names. Positive and missing Index values to be  *;
%*    postfixed to the prefix are at least one character in size, negative    *;
%*    Index values generate of a postfix of at least 2 characters (incl. '_').*;
%*    Duplicates are being checked and reported, after which the macro aborts.*;
%*   -If this argument is left empty the PreFxLst will be deduced from the    *;
%*    VarList automatically by taking the first 1 to 7 characters of each MR  *;
%*    variable name. The user should take care in this instance that these    *;
%*    significant parts of the names in the VarList are different! If they are*;
%*    not, or cause duplicate names, they are reported and the macro aborts.  *;
%*                                                                            *;
/* ')-Specific records of multiple records may be got lost in two instances:  */
%*    1. if the index value is outside the explicit user specified range,     *;
%*    2. if the index value is a duplicate within the same ByList values      *;
%*       combination, only the last (sorted) instance of duplicates is kept.  *;
%*    Both occurrences are reported to the log.                               *;
%*   -RM records (with missing values for all MR variables (but not CopyList  *;
%*    variables)) are also output for Index values outside the range lowest   *;
%*    to highest, with the number of missing records as a negative number and *;
%*    the number of passed records as 0. To exclude them in a later data step *;
%*    one should select either (_Missed_ GE 0) or (_Passed_ GE 1).            *;
%*                                                                            *;
%* )-DupIndex has default value Exclude. This is most logical and generally  *;
%*    valid. A user requested to also allow Include of duplicate Index values *;
%*    generating additional (preceding) records. This now is experimentally   *;
%*    implemented, but is not logical and generally should not be applied.    *;
%*    Additional records for each duplicate Index value are created. If       *;
%*    KeepVars=Yes then these also contain cumulative _Passed_, _Missed_      *;
%*    and _Record_ variables (_Record_= Outside | Duplicate | Standard).      *;
%*    With CopyList variables the values kept for each of them do not always  *;
%*    seem to be the most logical or desired ones. The UPDATE statement in    *;
%*    the code may complain by issuing warnings. Use Include with care!       *;
%*                                                                            *;
%* Variable name lists are supported as an explicit enumeration of common     *;
%* variable names. Implicit range lists are not (yet) supported and lists of  *;
%* non-standard variable names with embedded spaces are not supported either. *;
%*                                                                            *;
%* Output:                                                                    *;
%* =======                                                                    *;
%* A dataset (Out) in which all specified variables in the VarList are        *;
%* rearranged as (a list of) a needed number of differently named (Index      *;
%* numbered) variables within a single record per unique combination of       *;
%* (optionally formatted) values for ByList variables.                        *;
%* This can make a hierarchically structured database flat.                   *;
%*                                                                            *;
%* All variable attributes of the original MR variables are automatically     *;
%* copied to the resulting variables, including their labels. If an original  *;
%* variable has no label a standard label with the index value is supplied.   *;
%* If the number of original multiple records per ByList value (combination)  *;
%* varies this may cause missing values for higher numbered result variables. *;
%*                                                                            *;
%* Furthermore the resulting dataset contains the ByList variables by which   *;
%* it is explicitely sorted.                                                  *;
%* The single index variable is not contained in the result (its values are   *;
%* contained within the resulting RM variable names).                         *;
%* Optionally, if the argument FreqVars=Y the specific variables _Passed_ and *;
%* _Missed_ are also included in the result.                                  *;
%*                                                                            *;
%* Optionally too the resulting dataset may contain all unconverted variables *;
%* as well, either - with their values of the last record of a set of sorted  *;
%*   multiple records, assuming them constant within the set, or              *;
%* - with their last (possibly) non-missing values of the records of a set of *;
%*   sorted multiple records, assuming them unique or missing within the set. *;
%*    Warning: the valid application of the CopyList feature is up to the     *;
%*    responsibility of the user as values of previous records may be lost!   *;
%* Any other unwanted variables might be dropped by the user outside (before  *;
%* or after the call to) this macro.                                          *;
%*                                                                            *;
%* Examples:                                                                  *;
%* =========                                                                  *;
%* 1. conversion of a single variable (representing repeated measures):       *;
%*    - for records determined by the _unformatted_ values of one             *;
%*      identification variable (whether a format is associated or not),      *;
%*    - using the values of an existing index variable,                       *;
%*    - generating the resulting RM variable name automatically as RMvar#     *;
%*      (#=0 to 5, specified as Lowest and Highest),                          *;
%*    - while desiring to know the number of passed and missed multiple       *;
%*      records                                                               *;
/*    %MR2RM (Data=Input, Out=Output, ByList=CaseNr, FreqVars=Y,              */
/*            VarList=RMvar, Index=Repetitn, Lowest=0, Highest=5);            */
%*                                                                            *;
%* 2. conversion of an explicitely specified set of repeated measures         *;
%*    variables:                                                              *;
%*    - for records determined by the eventually _formatted_ values of two    *;
%*      identification variables (whether a format is associated or not),     *;
%*    - using all values of an automatically generated (temporary) index      *;
%*      variable,                                                             *;
%*    - specifying the prefixes for the resulting RM variables,               *;
%*    - keeping all possible variables, including the unspecified remaining   *;
%*      ones (with values of the last original record of a sorted case)       *;
/*    %MR2RM (Data=Input, Out=Output, ByList=GroupNr CaseNr,                  */
/*            VarList=VarA VarB VarC VarD VarE, CopyList=_REST_, CopyFrom=    */
/*            LastRec, ByFormat=, PreFxLst=AVar BVar CVar DVar EVar);         */
%*                                                                            *;
%* Remarks:                                                                   *;
%* ========                                                                   *;
%* The macro does not make use of PROC TRANSPOSE that works quite differently.*;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
%* = internal: - [none]                                                       *;
%* = attached: - [none]                                                       *;
%* = external: - ChkVar PreZeros FirstCh ChkList CrossChk                     *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of MR2RM:     (versions 0.x and 1.x developed as Y_MK_RM at IKR)   *;
%* =================                                                          *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0      14 May 1998  first working release converting a single variable   *;
%* 0.9       4 Jan 1999  made suitable for both alpha and numeric variables   *;
%* 1.0       5 Jan 1999  minor improvements and bug removals                  *;
%* 1.0.1    15 Feb 1999  added len (length of result variables)               *;
%* 1.0.2    16 Feb 1999  added some minor checks on arguments                 *;
%* 2.0      14 Jun 1999  made suitable for any number of MR variables at once *;
%* 2.1      14 Jun 1999  added automatic determination of all attributes      *;
%* 2.2      14 Jun 1999  made suitable for any number of BY variables at once *;
%* 2.2.1    15 Jun 1999  added automatic labelling if no labels, minor changes*;
%* 2.2.2    15 Jun 1999  added automatic deduction of extreme index values    *;
%* 2.2.3    15 Jun 1999  added automatic generation of index if none specified*;
%* 2.2.4    16 Jun 1999  added check for duplicate Index values within ByList *;
%*                       added recoding all 28 missing Index values to value .*;
%*                       added conversion of real Index values to integers    *;
%*                       added support for negative Index values (first time) *;
%*                       added support for missing  Index values (first time) *;
%* 2.2.5    16 Jun 1999  added optional counts of passed and missed MRs       *;
%* 2.2.6    18 Jun 1999  added automatic deduction of PreFxLst from VarList   *;
%* 2.2.7     2 Jul 1999  added code to compile but skip this macro further    *;
%* 2.2.8    20 Jul 1999  added %LOCAL statement for all local variables       *;
%* 2.2.9     2 Aug 1999  added Lowest as start of automatic index values,     *;
%*                       added check for existing Data dataset,               *;
%*                       added check for exist. vars in ByList, VarList, Index*;
%*                       added check for non-existing Out dataset,            *;
%*                       added OverWrit argument for existing Out dataset     *;
%*                       adapted source code layout to answer GPP guidelines  *;
%* 2.3.0     3 Aug 1999  changed varlist checking by calling external macro,  *;
%*                       replaced Version argument by ForMyUse argument,      *;
%*                       added KeepRest argument for keeping remaining vars   *;
%*  -       16 Aug 1999  added support for already formatted ByList variables,*;
%*                       minor improvements (layout, restricting variables,   *;
%*                          internal macro name)                              *;
%* 2.4.0    18 Aug 1999  added check on numeric type of index variable,       *;
%*                       added argument ByFormat to select the method for     *;
%*                          identifying cases from identification variables,  *;
%*                       improved user documentation above accordingly        *;
%* 2.4.1    24 Sep 1999  changed 'ABORT ABEND' into only 'ABORT' with err0rs  *;
%* 2.4.2    28 Dec 1999  fixed bug with empty index variable                  *;
%* 2.5.0    21 Jan 2000  corrected limitation to 9 MR variables by lifting it,*;
%*                       added forgotten retaining MR vars with missing Index,*;
%*  -       24 Jan 2000  inhibited creating RM vars for non-existing indexes, *;
%*                       added adding RM index number to existing MRvar label *;
%* 2.5.1    23 Feb 2000  fixed four bugs (partly to do with new feature)      *;
%* 2.5.2    13 Jun 2001  corrected check for numeric existing Index variable, *;
%*                       corrected forcing integer index values in list only  *;
%* 2.5.3    14 Jun 2001  added VarList value '_ALL_' or none for all variables*;
%* 2.5.4    15 Jun 2001  fixed bug generating RM labels with Index values     *;
%*                       fixed bug with VarList value _ALL_: ByList, Index    *;
%* 2.5.5    22 Jun 2001  removed ATTRIB LENGTH after SET, causing warning,    *;
%*                         where a text variable copy also sets correct length*;
%*                       disabled processing variable labels, bugs with quotes*;
%* 2.5.6    27 Jun 2001  fixed bug processing variable labels with quotes,    *;
%*                       forced index value in variables names maximum length,*;
%*                       changed adding '(#)' into adding '/#' to the label,  *;
%*                       added shortening too long labels at first (!) slash, *;
%*                       changed KeepRest argument into CopyList and CopyFrom *;
%* 2.6.0    29 Jun 2001  added check on equally specified or generated prefix,*;
%*                         and on unique, indexed generated RM variable names *;
%* 2.6.1     5 Jul 2001  added IndPreFx argument, Index prefix in var. labels *;
%* 2.6.2     8 Oct 2001  fixed bug causing all variables kept if index spec., *;
%*                       added no_index indication in label if IndPreFx empty *;
%* 2.6.3     9 Oct 2001  internal changes with application of NRBQUOTE funct. *;
%* 2.7.0    30 Oct 2001  undone 2.6.3 changes causing eating work disk space, *;
%*                       added varname and -label length for SAS V6 and V7/8  *;
%* 2.7.1     2 Nov 2001  fixed bug with empty IndPreFx (undeclared macrovar)  *;
%* 2.7.2    21 Nov 2001  changed reporting duplicate ByList and Index values  *;
%*                         into once per MR record, not for each MR variable, *;
%*                       changed counting _Passed_ and _Missed_ similarly     *;
%* 2.7.3a   15 Apr 2004  reenabled ATTRIB LENGTH (SAS 8.2), disabled in 2.5.5,*;
%*                       adapted PROC CONTENTS Name into UPCASE(Name) because *;
%*                          of SAS 8.2 backwards incompatibility with 6.12,   *;
%*                       changed calling %PreZeros with positional parameters *;
%*                          into calls with keyword parameters, also because  *;
%*                          of SAS 8.2 backwards incompatibility with 6.12,   *;
%*                       removed _LAST_ as default Data value (for later)     *;
%* 2.7.3b   14 Dec 2004  corrected bug with check on unique RM variable name, *;
%*                       corrected counting _Passed_ and _Missed_ for every   *;
%*                          unique ByList value combination, i.e. RM record   *;
%* 2.7.3c   15 Dec 2004  fixed bug generating RM vars outside low-high range, *;
%*                       changed CopyList keyword _ALL_ into _REST_           *;
%* 2.7.3d   17 Dec 2004  added CalledBy argument, Version and Debug control,  *;
%*                       added check on existence of CopyList variables,      *;
%*                       added aux. macro First to obtain first UC character  *;
%* 2.7.3e   23 Dec 2004  fixed bug preventing outputting RM records with      *;
%*                          Index values larger than specified highest,       *;
%*                       added counting of index outside range in _Missed_,   *;
%*                       added DupIndex argument to include duplicate Index) *;
%*                          and adapted counting _Passed_ and _Missed_        *;
%* 2.7.3f   24 Dec 2004  minor code changes, improvements and argument checks,*;
%*                       added aux. macro CrossChk and extensive list checks  *;
%*          27 Dec 2004  corrected bug with CopyList and DupIndex=Include,    *;
%*                       extended search for duplicate names with RMs created,*;
%*                       improved deletion of temporary work datasets         *;
%*          28 Dec 2004  corrected bugs with missing Index values and Index   *;
%*                          values outside range Lowest to Highest,           *;
%*                       added _Record_ output variable if FreqVars=Yes and   *;
%*                          DupIndex=Include to report record type,           *;
%*                       added forcing output of records with Index values    *;
%*                          outside range if FreqVars=Yes to report valid     *;
%*                          _Passed_, _Missed_ and _Record_ values,           *;
%*                       added forcing outputting Index values if DupIndex=I  *;
%*           4 Jan 2005  fixed/rewritten _Passed_/_Missed_ calculation code,  *;
%*                          see Decision table in code for various conditions,*;
%*                       if FreqVars=Yes and Index value never inside range   *;
%*                          then _Missed_ gets negative (no last RM to write) *;
%*                       fixed bugs with empty Lowest and Highest values      *;
%* 2.7.3g    5 Jan 2005  minor (textual) improvements, fixed minor bugs,      *;
%*                       added check on single Index variable specification,  *;
%*                       added check on FreqVars specification,               *;
%*                       added labels for _Passed_, _Missed_ and _Record_     *;
%* 2.7.3h    6 Apr 2005  enabled accidentally disabled temp ds deletion code, *;
%*                       added _LAST_ support with Data and Out dataset again *;
%*           2 Oct 2005  added CalledBy arg to Prezeros calls                 *;
%* 2.7.3i   21 Oct 2005  changed calls to internal macro First into external  *;
%*                         macro FirstCh and removed internal First code      *;
%*                       changed internal macros ChkList and CrossChk into    *;
%*                         external ones and removed their internal code      *;
%*           7 Nov 2005  redesigned normal and illegal termination actions,   *;
%*                       (re)set SYSLAST variable to last Data or Out         *;
%* 2.7.3j   21 Mar 2006  SAS names part for vs 9 set equal to vs 8 (for now)  *;
%* 2.7.3k   10 May 2006  several parameter name changes (aka proc TRANSPOSE)  *;
%*                       - changed parameter name SourceDs into Data    ,     *;
%*                       - changed parameter name TargetDs into Out     ,     *;
%*                       - changed parameter name IDlist   into ByList  ,     *;
%*                       - changed parameter name IDformat into ByFormat,     *;
%*                       - changed parameter name MRlist   into VarList ,     *;
%*                       - changed parameter name KeepList into CopyList,     *;
%*                       - changed parameter name KeepFrom into CopyFrom,     *;
%*                       (old names yet supported until some future version), *;
%*                       added QUIT after RUN with PROC DATASETS,             *;
%*                       changed tests on &ByFormat to search for whole word, *;
%*                       added support for all 28 missing Index values, which *;
%*                         generate unique RM variables for each of them,     *;
%*                       added WHERE argument to add a complete WHERE clause  *;
%* 2.7.3l   11 Jun 2007  prevented defining formats of only a $ character,    *;
%*                       corrected buggy WHERE argument (with PROC CONTENTS)  *;
%* 2.7.3m   28 Oct 2009  added IndexLen argument to specify the width (number *;
%*                         of digits) of the index postfix to the repeated    *;
%*                         variable name, dependent on the 'widest' index:    *;
%*                       - Maximum: index postfix width = width of longest    *;
%*                         index value in digits, constant for all indexes,   *;
%*                         smaller, narrower values are prefixed with zeros   *;
%*                       - Minimum: index postfix width = index digits width  *;
%*                         as is, no zero prefixes, may be variable width     *;
%* 2.7.3n   26 Jul 2012  fixed more bugs with empty Lowest and Highest values *;
%*                                                                            *;
%* Future features:                                                           *;
%* ================                                                           *;
%* - check Lower and Higher for numerical content.                            *;
%* - check dataset and variable names using my SAS_name macro.                *;
%* - support for (checking of) non-standard variable names ('..........'n).   *;
%* - generate dataset labels in the same way (from the original labels)/auxM. *;
%* - [ apply IdFrqLst like in in other macro(s) ??? ].                        *;
%* - better check on integer Lowest and Highest values, especially from var.. *;
%* - (instead of an Index variable: an Index value list, e.g. 2 4 7 10 ...).  *;
%* - VarList = _AUTO_ automatically deduce variables with multiple non-missing*;
%*      values to include in the list to restructure. (Like external VarList),*;
%*      excluding variables from ByList, Index, ExclList and CopyList.        *;
%* - ExclList argument to specify variables to exclude from the VarList.      *;
%* - MisVaLst argument: list of default (missing) values for VarList variables*;
%*      if multiple record(s) for index value(s) do(es) not exist for ByList, *;
%*      empty: . or space, single value: for all same types, list of values.  *;
%* - Listsep argument: element separator for all lists, esp. if MisvaLst=' '. *;
%* - an argument specifying a new explicit format list for ByList variables:  *;
%*   ByFormat extension: = fmt [fmt [...] ], specific formats for ByList,     *;
%*   !: for all ByList variables, #: for each separate ByList variable.       *;
%* - [ experiment with and check already formatted ByList variables, esp.     *;
%*      different unformatted values with identical FORMATTED values ].       *;
%* - a separate macro RM2MR doing the reverse, a prototype from 1998 exists.  *;
%* - store state of options MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN    *;
%*      to reset them at their original state at the end if changed to debug. *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS NOMPRINT NOMERROR NOSERROR NOMLOGIC NOSYMBOLGEN NOMACROGEN;
%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* remove comment chars for testing ;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN %GOTO Finish;

%IF (&Debug EQ 1 OR &Debug GT 11) %THEN %DO;
  OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* LOCAL macro variables declarations ;
%*____________________________________;

 %LOCAL    /* Define internal macro variables explicitely local */
  MacName  /* This macros external name: MR2RM */
  MacroVs  /* Version number major.minor.patch of macro: 2.7.3n */
  MacDate  /* Date of this version of macro: 26 Jul 2012 */
  I        /* element index */
  ByVar    /* ByList element */
  J        /* incremental value for several purposes */
  K        /* incremental value for several purposes */
  Var_List /* variable list (to check for existence or consistency) */
  ClasList /* all ByList elements eventually preceeded by "GROUPFORMAT" */
  LastBy   /* last ByVar from ByList */
  MultRec  /* VarList element */
  LabelLen /* length of current (MR) variable label */
  MrVarLbl /* Variable label space for MR variable label */
  PreFix   /* PreFxLst element or auto-deduced */
  PreFix2  /* PreFxLst element or auto-deduced, next */
  _Sorted1 /* stand-in name for sorted dataset 1 XX_Sort1 */
  _Sorted2 /* stand-in name for sorted dataset 2 XX_Sort2 */
  KeepVars /* stand-in name for kept variables dataset XX_KeepV */
  _Extreme /* stand-in name for Extremes dataset XX_Extr_ */
  Contents /* stand-in name for Contents dataset XX_Cont_ */
  IndexFrq /* stand-in name for actual index value list dataset */
  IndexLst /* list of actually occurring index values for variable */
  IndexVal /* element of IndexLst, one of the values for the Index */
  MaxIxLen /* maximum length of Index value */
  RMvarIx  /* resulting Repeated Measures variable name with Index */
  RMlstIx  /* resulting RM var name list with Index from current single MRvar */
  RMlist   /* complete Repeated Measures variable name list being generated */
  TooLong  /* variable label length plus index value indication length minus 40 */
  ErrCount /* number of erroneously specified arguments */
  Err      /* text 'E R R O R' as consecutive characters */
  Warn     /* text 'W A R N I N G' as consecutive characters */
  MaxName  /* maximum variable name length */
  MaxLabel /* maximum variable lebel length */
  ValidVar /* ValidVarName option value */
  ByIndex  /* indication of Index also specified as (part of) ByList */
  MRindex  /* indication of Index also specified as (part of) VarList */
  ToOutput /* indication of multiple record to be retained and output */
  ;

%* Stand-in variable and dataset names;
  %LET MacName  = MR2RM   ;
  %LET MacroVs  = 2.7.3n  ;
  %LET MacDate  = %STR(26 Jul 2012);
  %LET _Sorted1 = _YMRSrt1;
  %LET _Sorted2 = _YMRSrt2;
  %LET KeepVars = _YMRKpVr;
  %LET _Extreme = _YMRExtr;
  %LET Contents = _YMRCont;
  %LET IndexFrq = _YMRFreq;
  %LET ToOutput = _YMR2Out;
  %LET Err      = %SYSFUNC(COMPRESS(E R R O R, %STR( )));
  %LET Warn     = %SYSFUNC(COMPRESS(W A R N I N G, %STR( )));

%* Macro header in LOG file, dependent on value of CalledBy;
 %IF ( %UPCASE(&CalledBy) EQ %UPCASE(&MacName) OR &CalledBy EQ ) %THEN
 %DO;
  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs &MacroVs, by Jim Groeneveld, &MacDate   |;
  %PUT |      Multiple Records To Repeated Measures Conversion      |;
  %PUT |____________________________________________________________|;
  %PUT;
  
%* Process old parameter names into new ones;
  %LET ErrCount = 0;
%* Data;
  %IF (&Data NE AND &Data NE _LAST_ AND &SourceDs NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old SourceDs and new Data names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&SourceDs NE) %THEN %DO;
    %LET Data = &SourceDs; %* Data not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'SourceDs' extinct, use 'Data';
  %END;
%* Out;
  %IF (&Out NE AND &TargetDs NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old TargetDs and new Out names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&TargetDs NE) %THEN %DO;
    %LET Out = &TargetDs; %* Out not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'TargetDs' extinct, use 'Out';
  %END;
%* ByList;
  %IF (&ByList NE AND &IDlist NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old IDlist and new BYlist names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&IDlist NE) %THEN %DO;
    %LET ByList = &IDlist; %* BYlist not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'IDlist' extinct, use 'ByList';
  %END;
%* ByFormat;
  %IF (&ByFormat NE AND &IDformat NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old IDformat and new BYformat names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&IDformat NE) %THEN %DO;
    %LET ByFormat = &IDformat; %* BYformat not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'IDformat' extinct, use 'ByFormat';
  %END;
%* VarList;
  %IF (&VarList NE AND &MRlist NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old MRlist and new VarList names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&MRlist NE) %THEN %DO;
    %LET VarList = &MRlist; %* VarList not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'MRlist' extinct, use 'VarList';
  %END;
%* CopyList;
  %IF (&CopyList NE AND &KeepList NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old KeepList and new CopyList names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&KeepList NE) %THEN %DO;
    %LET CopyList = &KeepList; %* CopyList not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'KeepList' extinct, use 'CopyList';
  %END;
%* CopyFrom;
  %IF (&CopyFrom NE AND &KeepFrom NE) %THEN %DO;
      %PUT *** &MacName *** &ERR: Both old KeepFrom and new CopyFrom names specified;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF (&KeepFrom NE) %THEN %DO;
    %LET CopyFrom = &KeepFrom; %* CopyFrom not specified;
    %PUT *** &MacName *** &Warn: Parameter name 'KeepFrom' extinct, use 'CopyFrom';
  %END;
  %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

%* Feedback of specified arguments;
  %PUT &MacName: Parameter settings:;
  %PUT &MacName: Data (old: SourceDs) = &Data;
  %PUT &MacName: Where    = &Where   ;
  %PUT &MacName: Out  (old: TargetDs) = &Out;
  %PUT &MacName: OverWrit = &OverWrit;
  %PUT &MacName: ByList (old: IDlist) = &ByList  ;
  %PUT &MacName: ByFormat (old: IDformat) = &ByFormat;
  %PUT &MacName: VarList (old: MRlist) = &VarList  ;
  %PUT &MacName: PreFxLst = &PreFxLst;
  %PUT &MacName: Index    = &Index   ;
  %PUT &MacName: IndPreFx = %NRBQUOTE(&IndPreFx);
  %PUT &MacName: Lowest   = &Lowest  ;
  %PUT &MacName: Highest  = &Highest ;
  %PUT &MacName: IndexLen = &IndexLen;
  %PUT &MacName: FreqVars = &FreqVars;
  %PUT &MacName: CopyList (old: KeepList) = &CopyList;
  %PUT &MacName: CopyFrom (old: KeepFrom) = &CopyFrom;
  %PUT &MacName: DupIndex = &DupIndex;
  %PUT &MacName: Version  = &Version ;
  %PUT &MacName: Debug    = &Debug   ;
  %PUT &MacName: CalledBy = &CalledBy;
  %*PUT &MacName: ForMyUse = &ForMyUse;
 %END;

  %IF (&Version NE AND &Version NE &MacroVs) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Specified version &Version does not match &MacroVs,;
    %PUT .                       macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
    %GOTO Finish;
  %END;

%* Process keyword _LAST_ with both Data and Out;
  %IF (%UPCASE(&Data) EQ _LAST_) %THEN %LET Data = &SYSLAST;
  %IF (%UPCASE(&Out) EQ _LAST_) %THEN %LET Out = &SYSLAST;

  %IF ( %FirstCh(&DupIndex) NE E) %THEN
    %PUT &MacName: *** WARNING *** DupIndex=Include, see note in documentation;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Adaptation of variable name and label length limits to SAS version (6,7,8,9);
%*_____________________________________________________________________________;

  %IF (%SUBSTR(&SYSVER, 1, 2) EQ %STR(6.)) %THEN
  %DO; %* vs. 6.xx;
    %LET MaxName = 8;
    %LET MaxLabel = 40;
  %END;
  %ELSE %IF (%SUBSTR(&SYSVER, 1, 2) EQ %STR(8.)
          OR %SUBSTR(&SYSVER, 1, 2) EQ %STR(9.)) %THEN
  %DO; %* vs. 8.x.y, 9.x.y;
    %LET ValidVar = %UPCASE(%SYSFUNC(GETOPTION(VALIDVARNAME)));
    %IF (&ValidVar EQ V6) %THEN
    %DO;
      %LET MaxName = 8;
      %LET MaxLabel = 40;
    %END;
    %ELSE /* V7 or UPCASE */
    %DO;
      %LET MaxName = 32;
      %LET MaxLabel = 256;
    %END;
  %END;
  %ELSE
  %DO; %* any other vs.;
    %LET MaxName = 8;
    %LET MaxLabel = 40;
  %END;
%* How about format names in the various SAS versions 9.x? ;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Check arguments for legal contents and discrepancies, calling e_r_r_o_r_s    ;
*______________________________________________________________________________;

  %LET ErrCount = 0;
  %LET MRindex = 0;
  %LET ByIndex = 0;
  %IF ( &Lowest NE AND &Highest NE ) %THEN
  %DO;
    %IF ( &Lowest > &Highest ) %THEN
    %DO;
      %PUT *** &MacName *** &ERR: Lowest (&Lowest) larger than Highest (&Highest),;
      %PUT                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
    %END;
  %END;
/* @@@@@@@@@ check both &Lowest and &Highest for legal integer values here! */
  %IF ( &Data EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Data argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&Data)) EQ 0 ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Dataset Data=&Data does not exist,;
    %PUT                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE
  %DO;
    %IF ( %SCAN ( &Index, 2, %STR( ) ) NE ) %THEN /* standard name only */
    %DO;
      %PUT *** &MacName *** &ERR: More than one Index variables specified: Index=&Index,;
      %PUT .                      macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
    %END;
%* Exclusive checks on duplicate variable names within and between lists ;
      %ChkList  ( ListName=  ByList, VarList =  &ByList, Result=ErrCount);
      %CrossChk ( LstName1=   Index, VarList1=   &Index,
                  LstName2=  ByList, VarList2=  &ByList, Result=ByIndex);
    %IF (%UPCASE(&VarList) NE _ALL_) %THEN
    %DO;
      %ChkList  ( ListName=  VarList, VarList =  &VarList, Result=ErrCount);
      %CrossChk ( LstName1=  ByList, VarList1=  &ByList,
                  LstName2=  VarList, VarList2=  &VarList, Result=ErrCount);
      %CrossChk ( LstName1=   Index, VarList1=   &Index,
                  LstName2=  VarList, VarList2=  &VarList, Result=MRindex);
    %END;
    %IF (%UPCASE(&CopyList) NE _REST_) %THEN
    %DO;
      %ChkList  ( ListName=CopyList, VarList =&CopyList, Result=ErrCount);
     %IF (%UPCASE(&VarList) NE _ALL_) %THEN
      %CrossChk ( LstName1=  VarList, VarList1=  &VarList,
                  LstName2=CopyList, VarList2=&CopyList, Result=ErrCount);
      %CrossChk ( LstName1=  ByList, VarList1=  &ByList,
                  LstName2=CopyList, VarList2=&CopyList, Result=ErrCount);
      %CrossChk ( LstName1=  Index , VarList1=  &Index ,
                  LstName2=CopyList, VarList2=&CopyList, Result=ErrCount);
    %END;
%* Check for non-existing variables within dataset;
    %LET Var_List = &ByList &Index;
    %IF (%UPCASE(&VarList) NE %STR(_ALL_)) %THEN %LET Var_List = &Var_List &VarList;
    %IF (%UPCASE(&CopyList) NE %STR(_REST_)) %THEN %LET Var_List = &Var_List &CopyList;
    %ChkVar (Data=&Data, VarList=&Var_List, CalledBy=&MacName,
             Result = ErrCount); %* adding up non-existent variables ;
  %END;
  %IF ( %UPCASE(&Overwrit) NE YES AND %UPCASE(&Overwrit) NE NO
    AND %UPCASE(&Overwrit) NE Y AND %UPCASE(&Overwrit) NE N) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Overwrit argument not specified or illegal: &Overwrit,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( &Out EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Out argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&Out)) EQ 1 AND %FirstCh(&OverWrit) NE Y )
    %THEN %DO;
    %PUT *** &MacName *** &ERR: Dataset Out=&Out exists and OverWrit is not Y,;
    %PUT                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( &ByList EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: ByList argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
/*=========
  %IF ( &VarList EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: VarList argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
=========*/
  %IF ( %UPCASE(&ByFormat) NE FORMATTED AND %UPCASE(&ByFormat) NE UNFORMATTED
    AND %UPCASE(&ByFormat) NE F AND %UPCASE(&ByFormat) NE U) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: ByFormat argument not specified or illegal: &ByFormat,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( %UPCASE(&FreqVars) NE YES AND %UPCASE(&FreqVars) NE NO
    AND %UPCASE(&FreqVars) NE Y AND %UPCASE(&FreqVars) NE N) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: FreqVars argument not specified or illegal: &FreqVars,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( %UPCASE(&CopyFrom) NE LASTREC AND %UPCASE(&CopyFrom) NE UPDATED
    AND %UPCASE(&CopyFrom) NE L AND %UPCASE(&CopyFrom) NE U) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: CopyFrom argument not specified or illegal: &CopyFrom,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( %UPCASE(&DupIndex) NE EXCLUDE AND %UPCASE(&DupIndex) NE INCLUDE
    AND %UPCASE(&DupIndex) NE E AND %UPCASE(&DupIndex) NE I) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: DupIndex argument not specified or illegal: &DupIndex,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( %UPCASE(&IndexLen) NE MAXIMUM AND %UPCASE(&IndexLen) NE MINIMUM
    AND %UPCASE(&IndexLen) NE MAX AND %UPCASE(&IndexLen) NE MIN) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: IndexLen argument not specified or illegal: &IndexLen,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ChkList (Element=Prefix specification, ListName=PreFxLst, VarList=&PreFxLst, Result=ErrCount);
  %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Determine last ByVar variable from ByList and build Class list with "GROUPFORMAT" ;
%*___________________________________________________________________________________;

  %IF ( %UPCASE(&ByFormat) EQ F ) %THEN %LET ByFormat = FORMATTED;
  %ELSE %IF ( %UPCASE(&ByFormat) EQ U ) %THEN %LET ByFormat = UNFORMATTED;
  %IF ( %UPCASE(&ByFormat) NE UNFORMATTED ) %THEN %LET ClasList = ;
  %LET I = 1;
  %LET ByVar = %SCAN(&ByList,&I,%STR( ));
  %DO %WHILE (&ByVar NE);
    %LET LastBy = &ByVar;
    %IF ( %UPCASE(&ByFormat) NE UNFORMATTED ) %THEN
      %LET ClasList = &ClasList GROUPFORMAT &ByVar;
    %LET I = %EVAL( &I + 1 );
    %LET ByVar = %SCAN(&ByList,&I,%STR( ));
  %END; %*=== While;

  %IF ( %UPCASE(&ByFormat) EQ UNFORMATTED ) %THEN %LET ClasList = &ByList;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Build VarList from keyword _ALL_, exclude ByList and Index variables ;
%*______________________________________________________________________;

  %IF (%UPCASE(&VarList) EQ %STR(_ALL_) OR %UPCASE(&VarList) EQ /*empty*/) %THEN
  %DO;
    %LET VarList = ; %* empty;
    PROC CONTENTS DATA=&Data (DROP=&ByList &Index
        %IF (%UPCASE(&CopyList) NE _REST_) %THEN &CopyList;
                  ) OUT=&Contents (KEEP=Name) NOPRINT;
    RUN;

    DATA _NULL_;
      SET &Contents;
      CALL EXECUTE ( '%LET VarList = &VarList ' || Name || ';' );
    RUN;
  %PUT &MacName: Generated VarList=&VarList;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Build CopyList from keyword _REST_, exclude ByList, VarList and Index var.s ;
%*_____________________________________________________________________________;

  %IF (%UPCASE(&CopyList) EQ %STR(_REST_)) %THEN
  %DO;
    %LET CopyList = ; %* empty;
    PROC CONTENTS DATA=&Data (DROP=&ByList &Index &VarList)
                  OUT=&Contents (KEEP=Name) NOPRINT;
    RUN;

    DATA _NULL_;
      SET &Contents;
      CALL EXECUTE ( '%LET CopyList = &CopyList ' || Name || ';' );
    RUN;
  %PUT &MacName: Generated CopyList=&CopyList;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic determination of attributes and check on numeric index variable ;
%*___________________________________________________________________________;

  PROC CONTENTS DATA=&Data (KEEP=&VarList &Index)
                 OUT=&Contents (KEEP=Name Label Informat Format Type Length)
                 NOPRINT;
  RUN;
* dbg;
/*proc print data=&contents; run;*/

  DATA _NULL_;
    SET &Contents;

    %LET I = 1;
    %LET MultRec = %SCAN(&VarList,&I,%STR( ));
    %DO %WHILE (&MultRec NE);

      %IF ( &I GT 1 ) %THEN ELSE;
      IF ( UPCASE(Name) EQ UPCASE("&MultRec") ) THEN
      DO;
%* Make attributes as macro variables;
       %LOCAL MR&I.L MR&I.T MR&I.B MR&I.F MR&I.I; %* Define locals;
        CALL SYMPUT ("MR&I.L", TRIM(LEFT(PUT(Length,3.0))));
        CALL SYMPUT ("MR&I.T", TRIM(LEFT(PUT(Type,1.0))));
        CALL SYMPUT ("MR&I.B", TRIM(LEFT(Label)));
        CALL SYMPUT ("MR&I.F", TRIM(LEFT(Format)));
        CALL SYMPUT ("MR&I.I", TRIM(LEFT(InFormat)));
      END;

      %LET I = %EVAL( &I + 1 );
      %LET MultRec = %SCAN(&VarList,&I,%STR( ));
    %END; %*=== While;

    %IF ( &Index NE ) %THEN
    %DO;
      IF ( UPCASE(Name) EQ UPCASE("&Index") ) THEN
      DO;
        IF ( Type NE 1 ) THEN
        DO;
          PUT "*** &MacName *** &ERR: Index variable &Index is not numeric,";
          PUT "                     macro &MacName will abort";
          CALL SYMPUT ("ErrCount", "1");
        END;
      END;
    %END;
  RUN;

/*=====
  PROC DATASETS NOLIST;
    DELETE &Contents;                        %* Delete scratch dataset &Contents;
  RUN; QUIT;
=====*/

  %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic generation of index variable (if none specified) ;
%*____________________________________________________________;

  %IF ( &Index EQ ) %THEN
  %DO;
    %LET Index = XX_Index; %* XX_Index assumed unique temporary name;
    PROC SORT DATA=&Data (KEEP=&ByList &VarList &CopyList
      %IF (%QUOTE(&Where) NE) %THEN WHERE = (&Where);
             ) OUT=&_Sorted1; BY &ByList;   %* Sort input file;
     %IF ( %UPCASE(&ByFormat) EQ UNFORMATTED ) %THEN %STR(FORMAT &ByList;) ;
    RUN;
    DATA &_Sorted1;
      SET &_Sorted1; BY &ClasList;
      IF ( FIRST.&LastBy ) THEN &Index = %IF ( &Lowest NE ) %THEN &Lowest; %ELSE 1;;
      ELSE &Index + 1;
    RUN;
  %END;
  %ELSE
  %DO; %* Just copy original source dataset, while KEEPing specific variables;
    DATA &_Sorted1;
      SET &Data (KEEP=&ByList &VarList &Index &CopyList
          %IF (%QUOTE(&Where) NE) %THEN WHERE = (&Where); );
    RUN;
  %END; %* End - reverse of ( &Index EQ ) ;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic determination of minimum and maximum occurring Index values ;
%*_______________________________________________________________________;

  %IF ( &Lowest EQ OR &Highest EQ ) %THEN
  %DO;
    PROC UNIVARIATE DATA=&_Sorted1 (KEEP=&Index) NOPRINT;
      VAR &Index;
      OUTPUT OUT = &_Extreme
             MIN = Minimum
             MAX = Maximum ;
    RUN;

    DATA _NULL_;
      SET &_Extreme;
      %IF ( &Lowest  EQ ) %THEN CALL SYMPUT ("Lowest" , TRIM(LEFT(PUT(Minimum,9.0))));;
      %IF ( &Highest EQ ) %THEN CALL SYMPUT ("Highest", TRIM(LEFT(PUT(Maximum,9.0))));;
    RUN;

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
%* If &Lowest and/or &Highest still are empty due to no extreme data ;
%* then define them as missing (._ and .z). [26jul2012, vs. 2.7.3n]  ;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
    %IF ( &Lowest  EQ ) %THEN %LET Lowest  = ._; %* smallest missing value;
    %IF ( &Highest EQ ) %THEN %LET Highest = .z; %* largest  missing value;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Convert eventual real Lowest and Highest values to integers and set max len ;
%*_____________________________________________________________________________;

  %IF ( &Lowest  GT .z ) %THEN %LET Lowest  = %SYSEVALF(&Lowest ,CEIL);
  %IF ( &Highest GT .z ) %THEN %LET Highest = %SYSEVALF(&Highest,CEIL);
  %LET MaxIxLen = %SYSFUNC(MAX(%LENGTH(&Lowest),%LENGTH(&Highest)));
%*put @@@@@@@@@26jul2012@@@@@@@@@ [Lowest=&Lowest, Highest=&Highest];

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Make list of really occurring index values within explicit range ;
%*__________________________________________________________________;

  PROC FREQ DATA=&_Sorted1;
    FORMAT &Index BEST8.0; %* Force freqs of integer values;
    TABLES &Index / NOPRINT NOPERCENT OUT=&IndexFrq; RUN;

  %LOCAL IndexVal Standard Special; %* Variable names;
  %LET IndexVal = ___IxVal;
  %LET Standard = ___Stand;
  %LET Special  = ___Spec;
  DATA _NULL_;
    LENGTH &IndexVal $ 8;
    RETAIN &Standard &Special;
    SET &IndexFrq END=LastRec;
    &IndexVal = LEFT(PUT(&Index,8.));
*** IF (&Index LE .z) THEN &IndexVal = '.'; %* or OPTIONS MISSING='.' + reset;
    IF (&Index EQ .) THEN &Standard = 1;
    ELSE IF (&Index LE .z) THEN DO; /* 9 May 2006 */
      &IndexVal = '.' || &IndexVal;
      IF (&Index EQ ._) THEN &Special = 1;
    END;
    IF (&Index LE .z OR (&Index GE &Lowest AND &Index LE &Highest)) THEN
    /* see 15 Dec 2004 */
      CALL EXECUTE ( '%LET IndexLst = &IndexLst ' || &IndexVal || ';' );
    IF (LastRec AND &MaxIxLen EQ 1 AND &Standard AND &Special) THEN
      CALL SYMPUT ('MaxIxLen', '2');
  RUN;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check lowest and highest values ;
%*_________________________________;

  %IF ( &Lowest > &Highest ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Lowest (&Lowest) larger than Highest (&Highest),;
    %PUT                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
    %GOTO Aborting;
  %END;
  %ELSE %PUT &MacName: Lowest index value=&Lowest and Highest index value=&Highest;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic deduction of PreFxLst (if none specified) ;
%*_____________________________________________________;

  %IF ( &PreFxLst EQ ) %THEN
  %DO;

    %LET I = 1;
    %LET MultRec = %SCAN(&VarList,&I,%STR( ));
    %DO %WHILE (&MultRec NE);

      %IF ( %LENGTH(&MultRec) LE %EVAL(&MaxName-&MaxIxLen) ) %THEN %LET PreFix = &MultRec;
      %ELSE %LET PreFix = %SUBSTR ( &MultRec, 1, %EVAL(&MaxName-&MaxIxLen) );
      %LET PreFxLst = &PreFxLst &PreFix; %* Build PreFxLst;

      %LET I = %EVAL( &I + 1 );
      %LET MultRec = %SCAN(&VarList,&I,%STR( ));
    %END; %*=== While;

  %END; %* PreFxLst empty;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro actually starts here. Sort Source ;
%*_________________________________________;

  PROC SORT DATA=&_Sorted1 OUT=&_Sorted2; BY &ByList &Index;
    %IF ( %UPCASE(&ByFormat) EQ UNFORMATTED ) %THEN %STR(FORMAT &ByList;) ;
    FORMAT &Index;             %* Dissociate eventual format from index variable;
  RUN;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Start creation of Target (and possibly KeepVars with kept variables) ;
%*______________________________________________________________________;

  DATA &Out (DROP=&VarList &ToOutput
   %IF (NOT (&ByIndex OR %FirstCh(&DupIndex) EQ I)) %THEN &Index;)
/* No {(KEEP=&ByList &CopyList &RMlist)} because &RMlist does not yet exist */

   %IF (&CopyList NE AND %FirstCh(&CopyFrom) EQ U) %THEN
   %DO;
     &KeepVars (KEEP=&ByList &CopyList)
   %END;
   ;
    SET &_Sorted2; BY &ClasList &Index; %* Sorted input file necessary for LAST.BY;
   %IF (&CopyList NE AND %FirstCh(&CopyFrom) EQ U) %THEN
   %DO;
    IF (&Index LE .z OR (&Index GE &Lowest AND &Index LE &Highest)) THEN
      OUTPUT &KeepVars; * Produce scratch dataset with variables to keep;
   %END;

    IF ( &Index GT .z ) THEN &Index = CEIL (&Index); %* Make index integer;
*** ELSE &Index = .; %*Reduce all different missing values to one;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Manipulate each of the Multiple Record variables one by one separately ;
%*________________________________________________________________________;

    RETAIN &ToOutput 0;

   %LET RMlist = ;
   %LET I = 1;
   %LET PreFix = %SCAN(&PreFxLst,&I,%STR( ));
   %DO %WHILE (&PreFix NE);
     %LET MultRec = %SCAN(&VarList,&I,%STR( ));

     %IF ( &MultRec EQ ) %THEN
     %DO;
       %PUT *** &MacName *** &ERR: Prefix List longer than Multiple Records List;
       %PUT                      macro &MacName will abort; /* calling err */
       %LET ErrCount = %EVAL ( &ErrCount + 1);
       %GOTO Aborting;             %* @@@@@@@@@ allow this!!!;
     %END;

%* Recode macro variable &&MR&I.T and .L;
     %IF ( &&MR&I.T = 2 ) %THEN %LET MR&I.T = $; %ELSE %LET MR&I.T = ; %* empty;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check generated RM variable names against existing variable names firstly  ;
%*----------------------------------------------------------------------------;
%* Retain values (within unique ByList value combinations) and set attributes ;
%*----------------------------------------------------------------------------;
%* For each Multiple Record variable define attributes for each Index value   ;
%*____________________________________________________________________________;

* Change FixedLen for macro PreZeros into 0 (variable) if IndexLen=Minimum;
     %IF (%UPCASE(%SUBSTR(&IndexLen,1,3)) EQ MIN) %THEN %LET MaxIxLen = 0;

     %LET RMlstIx = ;
     %LET J = 1 ;
     %LET IndexVal = %SCAN ( &IndexLst, &J, %STR( ) );
     %DO %WHILE (&IndexVal NE);
       %LET RMvarIx=&PreFix.%PreZeros(Original=&IndexVal,FixedLen=&MaxIxLen,
                                      Calledby=&MacName);

%* Check generated RM variable names against existing variable names firstly  ;
       %LET K = 1;
       %LET ByVar = %SCAN(&ByList &VarList &Index &CopyList &RMlist &RMlstIx,
                       &K, %STR( )); %* added above+below: &RMlstIx [27dec2004];
       %DO %WHILE (&ByVar NE);

         %IF (%UPCASE(&RMvarIx) EQ %UPCASE(&ByVar)) %THEN
         %DO;
           %PUT *** &MacName *** &ERR: RM variable name &RMvarIx equals existing or;
           %PUT              other generated variable name. Macro &MacName will abort;
           %LET ErrCount = %EVAL ( &ErrCount + 1);
         %END;

         %LET K = %EVAL( &K + 1 );
         %LET ByVar = %SCAN(&ByList &VarList &Index &CopyList &RMlist &RMlstIx,
                         &K, %STR( ));
       %END; %*=== While;
       %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

       %LET RMlstIx = &RMlstIx &RMvarIx; %* Build RM list from current single MR;
%* Retain values (within unique ByList value combinations) and set attributes ;
        RETAIN &RMvarIx;
%* For each Multiple Record variable define attributes for each Index value   ;
       ATTRIB &RMvarIx LABEL =
       %IF ( "%NRBQUOTE(&&MR&I.B)" NE " " ) %THEN
       %DO;
         %IF (%NRBQUOTE(&IndPreFx) NE) %THEN
         %DO;
%*put dbg &MacName: LABEL, I=&I, PreFix=&PreFix, MultRec=&MultRec, J=&J, IndexVal=&IndexVal;
           %LET LabelLen = %LENGTH(%NRBQUOTE(&&MR&I.B));
           %LET TooLong  = %SYSFUNC(MAX(0, %EVAL
             (&LabelLen + %LENGTH(&IndexVal) + %LENGTH(%NRBQUOTE(&IndPreFx)) - &MaxLabel)));
%*put dbg &MacName: LABEL, LabelLen=&LabelLen, TooLong=&TooLong IndPreFx=%NRBQUOTE(&IndPreFx);
           %IF (&TooLong GE &MaxLabel) %THEN
             %LET MrVarLbl = %NRBQUOTE(&&MR&I.B); %* Leave unchanged;
           %ELSE %IF (&TooLong EQ 0) %THEN  /* Add to original */
             %LET MrVarLbl = %NRBQUOTE(&&MR&I.B)%NRBQUOTE(&IndPreFx)&IndexVal;
           %ELSE           /* remove TooLong characters from the beginning */
             %LET MrVarLbl = %SUBSTR(%NRBQUOTE(&&MR&I.B),%EVAL(&TooLong+1))%NRBQUOTE(&IndPreFx)&IndexVal;
         /* "%NRBQUOTE(&MrVarLbl)" with ATTRIB causes all UPPER CASE, SAS bug? */
            "&MrVarLbl"
         %END;
         %ELSE "&&MR&I.B"; %* IndPreFx is empty: leave varlabel unchanged;
       %END;
       %ELSE "&MultRec repeated measure &Index=&IndexVal";
/**/     %IF ( &&MR&I.L NE ) %THEN LENGTH   = &&MR&I.T &&MR&I.L; /**/
       %IF ( &&MR&I.F NE AND &&MR&I.F NE $ ) %THEN FORMAT   = &&MR&I.F..;
       %IF ( &&MR&I.I NE AND &&MR&I.I NE $ ) %THEN INFORMAT = &&MR&I.I..; ;


%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Assign MR value to corresponding RM variable (with Index value index) ;
%*_______________________________________________________________________;

        IF ( &IndexVal EQ &Index ) THEN
        DO;
          &RMvarIx = &MultRec;
          &ToOutput = 1;
        END;

       %LET J = %EVAL( &J + 1 );
       %LET IndexVal = %SCAN ( &IndexLst, &J, %STR( ) );
     %END; %*=== While (processing all Index values)===;

%* Build total generated varname list upto and incl previous MR-variable;
%* Previous RM-variables of same MR-variable always are differing from current;
     %LET RMlist = &RMlist &RMlstIx;

     %LET I = %EVAL( &I + 1 );
     %LET PreFix = %SCAN(&PreFxLst,&I,%STR( ));
   %END; %*=== While (processing all MR variables)===;

   %IF ( %SCAN(&VarList,&I,%STR( )) NE ) %THEN
   %DO;             %* &I is one more than number of Prefixes;
     %PUT *** &MacName *** &ERR: Prefix List shorter than Multiple Records List;
     %PUT                      macro &MacName will abort; /* calling err */
     %PUT                     in size, macro &MacName will abort;
     %LET ErrCount = %EVAL ( &ErrCount + 1);
     %GOTO Aborting;
   %END;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Decision table for FreqVars, Index, DupIndex, LAST.&LastBy, LAST.&Index      ;
*______________________________________________________________________________;
*                                                                              ;
* Index       | outside range | inside range &lowest to &highest plus missings ;
*-------------+---------------+-----------------------+------------------------;
* DupIndex    | E |  Include  | Exclude               | Include                ;
*-------------+---+-----+-----+---------------+-------+---------------+--------;
* LAST.LastBy | L | NOT | LAST|  N    O    T  |  LAST |  N    O    T  |  LAST  ;
*-------------+---+---+-+-----+-------+-------+-------+-------+-------+--------;
* LAST.Index  | L | N | LAST  |  NOT  |  LAST |  LAST |  NOT  |  LAST |  LAST  ;
*-------------+---+---+---+---+-------+-------+-------+-------+-------+--------;
* FreqVars=Y: |   |   |   |   |       |       |       |       |       |        ;
* _Passed_    | . | . | . | . |   .   |   +1  |   +1  |  =+1  |  =+1  |   =+1  ;
* _Missed_    |+1 |=+1|=+1|=+1|  +1   |   .   |(-_M_) |(-_M_) |(-_M_) | (-_M_) ;
* _Record_    | . | =Outside= |   .   |   .   |   .   |Duplica|S t a n d a r d ;
*-------------+---+---+---+---+-------+-------+-------+-------+-------+--------;
* If ToOutput |oRM| output(RM)|       |       |output |output |output | output ;
* or FreqVars |.if| incl.dups |   .   |   .   |RM, no |(RM) + |(RM) + | RM, no ;
* = Yes       |=No| + outside |       |       |dups   |dups   |dups   | dups   ;
*-------------+---+---+---+---+-------+-------+-------+-------+-------+--------;
* Reset _V_=0 | Y | Y | Y | Y |   N   |   N   |   Y   |   Y   |   Y   |    Y   ;
*-------------+---+---+---+---+-------+-------+-------+-------+-------+--------;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Check whether Index values out of range ;
*_________________________________________;

    IF ((&Index GT .z AND &Index LT &Lowest) OR &Index GT &Highest) THEN
    DO; * if Index value outside range;
      PUT "*** &MacName *** &WARN: &Index value " &Index
          "out of range &Lowest...&Highest for &ByList value(s) " &ByList;
      %IF ( %FirstCh(&FreqVars) EQ Y ) %THEN
      %DO;
        _Missed_ + 1; /* both if DupIndex=Exclude and Include */
        %IF ( %FirstCh(&DupIndex) EQ I ) %THEN _Record_ = 'Outside  ';;
      %END;
      %IF (&Debug EQ 2) %THEN PUT "Debug outside range: " _N_= _passed_= _Missed_=;;
%* MR record will be skipped by comparing &IndexLst and &Index and no copy;
    END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check for duplicate Index values ;
%*__________________________________;

    ELSE IF ( NOT LAST.&Index AND "%FirstCh(&DupIndex)" EQ 'E' ) THEN
    DO;
      PUT "*** &MacName *** &WARN: Duplicate &Index value " &Index
          "for &ByList value(s) " &ByList;
      PUT "    Repeated measures values of previous record lost.";
      %IF ( %FirstCh(&FreqVars) EQ Y ) %THEN
        _Missed_ + 1;;
      %IF (&Debug EQ 2) %THEN PUT "Debug duplicate: " _N_= _passed_= _Missed_=;;
    END;
    ELSE /*--- IF ( LAST.&Index OR "%FirstCh(&DupIndex)" EQ 'I' ) THEN ---*/
    DO;
      %IF ( %FirstCh(&FreqVars) EQ Y ) %THEN
      %DO;
        _Passed_ + 1;
        %IF ( %FirstCh(&DupIndex) EQ I ) %THEN
        %DO;
          IF ( NOT LAST.&Index ) THEN _Record_ = 'Duplicate';
          ELSE /* IF ( LAST.&LastIndex ) THEN */ _Record_ = 'Standard ';
        %END;
      %END;
      %IF (&Debug EQ 2) %THEN
        PUT "Debug Last.Index or DupIndex=I: " _N_= _Passed_= _missed_=;;
    END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Process final MR record within ByList combination, write retained values ;
%*--------------------------------------------------------------------------;
%* Reset all _Passed_, _Missed_ and _Record_ values if FreqVars=Y           ;
%*--------------------------------------------------------------------------;
%* Reset all RM variables to missing, necessary in case a certain           ;
%* Index value does not occur for a subsequent ByList combination           ;
%*__________________________________________________________________________;

    IF ( LAST.&LastBy OR "%FirstCh(&DupIndex)" EQ 'I' ) THEN
    DO;
      %IF ( %FirstCh(&FreqVars) EQ Y ) %THEN
      %DO;
        IF (NOT &ToOutput) THEN _Missed_ = -_Missed_; * negative if outside range;
        LABEL _Passed_='Number of used MRecords for RM record'
              _Missed_='Number of unused MRecords (duplicates, outside range)';
        %IF ( %FirstCh(&DupIndex) EQ I ) %THEN
          LABEL _Record_='Type of MRecord: Standard / Duplicate / Outside range';;
      %END;
      IF ( &ToOutput OR "%FirstCh(&FreqVars)" EQ 'Y' ) THEN
      DO; /* also output records with Index outside range if FreqVars=Yes */
        OUTPUT &Out;  %* Output current final &ByList record and continue;
        &ToOutput = 0;
      END;
      %IF ( %FirstCh(&FreqVars) EQ Y ) %THEN
      %DO;                                      * Initialize/reset counters;
        _Passed_ = 0;
        _Missed_ = 0;
      %END;
      %LET I = 1;
      %LET PreFix = %SCAN(&PreFxLst,&I,%STR( ));
      %DO %WHILE (&PreFix NE);
        %LET J = 1 ;
        %LET IndexVal = %SCAN ( &IndexLst, &J, %STR( ) );
        %DO %WHILE (&IndexVal NE);  %* Reset for next &ByList to undo RETAIN;
          %LET RMvarIx=&PreFix.%PreZeros(Original=&IndexVal,FixedLen=&MaxIxLen,
                                         Calledby=&MacName);
          &RMvarIx =
            %IF ( &&MR&I.T = $ ) %THEN ' '; %ELSE . ; ;
          %LET J = %EVAL( &J + 1 );
          %LET IndexVal = %SCAN ( &IndexLst, &J, %STR( ) );
        %END; %*=== While;
        %LET I = %EVAL( &I + 1 );
        %LET PreFix = %SCAN(&PreFxLst,&I,%STR( ));
      %END; %*=== While;
    END;                               %* If not LAST.&LastBy then no OUTPUT;

  RUN; * === End of Target dataset generation ===;

  %IF (&CopyList NE AND %FirstCh(&CopyFrom) EQ U) %THEN
  %DO;
* Update master with (multiple) transaction records from KeepVars all at once;
    DATA &Out;
      UPDATE &Out &KeepVars; BY &ClasList; * This is the core of the update;
    RUN;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Process error conditions ;
%*__________________________;

  %Aborting:

  PROC DATASETS NOLIST;           %* Delete scratch datasets;
    %IF (%SYSFUNC(EXIST(&_Sorted1))) %THEN DELETE &_Sorted1;;
    %IF (%SYSFUNC(EXIST(&_Sorted2))) %THEN DELETE &_Sorted2;;
    %IF (%SYSFUNC(EXIST(&KeepVars))) %THEN DELETE &KeepVars;;
    %IF (%SYSFUNC(EXIST(&_Extreme))) %THEN DELETE &_Extreme;;
    %IF (%SYSFUNC(EXIST(&Contents))) %THEN DELETE &Contents;;
    %IF (%SYSFUNC(EXIST(&IndexFrq))) %THEN DELETE &IndexFrq;;
  RUN; QUIT;

  %IF ( &ErrCount GT 0 ) %THEN
  %DO;
    %IF (%SYSFUNC(EXIST(&Out)) AND
         %FirstCh(&OverWrit) EQ Y) %THEN %DO;
      PROC DATASETS; DELETE &Out; RUN; QUIT;
    %END;
    %LET SYSLAST=&Data; * reset to original SYSLAST value;
    DATA _NULL_;
      ABORT /*ABEND*/;
    RUN;
  %END;
  %ELSE
    %LET SYSLAST=&Out; * reset to explicit SYSLAST value;

  %Finish:

%**********;
%MEND MR2RM; * ----------------------end-of-macro-MR2RM------------------------;
***********;

